
import unittest
from pydocmd.document import Section


class PreprocessorTest(unittest.TestCase):

  docstring_from = 'foo'
  trim_trailing_whitespace = False
  preprocessor = None
  code = None
  expected_docs = None

  def test_preprocessor(self):
    if type(self) is PreprocessorTest:
      raise unittest.SkipTest('Base class')

    scope = {}
    exec(compile(self.code, '<string>', 'exec'), scope)
    docstring = scope[self.docstring_from].__doc__

    section = Section(None)
    section.content = docstring
    self.preprocessor.preprocess_section(section)

    got = section.content.strip()
    expected = self.expected_docs.strip()
    if self.trim_trailing_whitespace:
      got = self._trim_trailing_whitespace(got)
      expected = self._trim_trailing_whitespace(expected)

    print(self.preprocessor)
    assert got == expected

  @staticmethod
  def _trim_trailing_whitespace(s):
    lines = s.split('\n')
    return '\n'.join(x.rstrip() for x in lines)



# Copyright (c) 2017  Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""
This module implements preprocessing Markdown-like docstrings and converts
it to fully markdown compatible markup.
"""

import re


class Preprocessor(object):
  """
  This class implements the basic preprocessing.
  """

  def __init__(self, config):
    self.config = config

  def preprocess_section(self, section):
    """
    Preprocess the contents of *section*.
    """
    codeblock_opened = False
    current_section = None
    content = self._preprocess_refs(section)
    lines = []
    for line in content.split('\n'):
      if line.startswith("```"):
        codeblock_opened = (not codeblock_opened)
      if not line:
        current_section = None
      elif not codeblock_opened:
        line, current_section = self._preprocess_line(line, current_section)
      lines.append(line)
    section.content = '\n'.join(lines)

  def _preprocess_line(self, line, current_section):
    match = re.match(r'# (.*)$', line)
    if match:
      current_section = match.group(1).strip().lower()
      line = re.sub(r'# (.*)$', r'__\1__\n', line)

    # TODO: Parse type names in parentheses after the argument/attribute name.
    if current_section in ('arguments', 'parameters'):
      style = r'- __\1__:\3'
    elif current_section in ('attributes', 'members', 'raises'):
      style = r'- `\1`:\3'
    elif current_section in ('returns',):
      style = r'`\1`:\3'
    else:
      style = None
    if style:
      #                  | ident  | types     | doc
      line = re.sub(r'\s*([^\\:]+)(\s*\(.+\))?:(.*)$', style, line)

    return line, current_section

  def _preprocess_refs(self, section):
    """
    Parses references and replaces them with markdown links
    Syntax is:
        `#anchor` for in page reference
        `#::mod[.submod][#item][+n]` where
            `mod.submod` is the module, `item` is the module member and n is
            the number of duplicates in case of headers: markdown conflicts.
            You can also references class members: `#::mod.submod.cls#method`
    """
    def handler(match):
      mod = match.group('mod')
      ref = match.group('ref')
      parens = match.group('parens') or ''
      dup = match.group('dup')
      has_trailing_dot = False
      if not parens and ref.endswith('.'):
        ref = ref[:-1]
        has_trailing_dot = True

      if '#' in ref:
        text = ref.split('#')[-1]
        title = mod_ref = ref.replace('#', '.')
        anchor = text.lower()
      else:
        title = text = anchor = ref
        anchor = ref.replace('.', '')
        mod_ref = ref

      if self.config['headers'] == 'html':
          anchor = mod_ref
          if not mod and not anchor.startswith(section.identifier):
              anchor = section.identifier + '.' + anchor
      elif dup:
          anchor += '_' + dup[1:]

      result = '`{}`'.format(text + parens)
      link = self.link_lookup.get(mod_ref, self.link_lookup.get(ref))
      if mod and link:
        result = ('[' + result + '](' + link + '#' + anchor + ' "' + title + '")')
      else:
        result = '[' + result + '](#' + anchor + ')'
      if has_trailing_dot:
        result += '.'
      return (match.group('prefix') or '') + result
    return re.sub('(?P<prefix>^| |\t)#(?P<mod>::)?(?P<ref>[\w\d\._\#]+)(?P<parens>\(\))?(?P<dup>\+\d+)?', handler, section.content)



# Copyright (c) 2017  Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""
This module implements preprocessing Markdown-like docstrings and converts
it to fully markdown compatible markup.
"""

import re
import textwrap

class Preprocessor(object):
  """
  This class implements the preprocessor for restructured text.
  """
  def __init__(self, config=None):
    self.config = config

  def preprocess_section(self, section):
    """
    Preprocessors a given section into it's components.
    """
    lines = []
    in_codeblock = False
    keyword = None
    components = {}

    line = textwrap.dedent(section.content)

    for line in section.content.splitlines():
      line = line.rstrip()

      if line.startswith("```"):
        in_codeblock = not in_codeblock

      line_codeblock = line.startswith('    ')

      if not in_codeblock and not line_codeblock:
        match = re.match(r'\s*:(?:param|parameter)\s+(\w+)\s*:(.*)?$', line)
        if match:
          keyword = 'Arguments'
          param = match.group(1)
          text = match.group(2)
          text = text.strip()

          component = components.get(keyword, [])
          component.append('- `{}`: {}'.format(param, text))
          components[keyword] = component
          continue

        match = re.match(r'\s*:(?:return|returns)\s*:(.*)?$', line)
        if match:
          keyword = 'Returns'
          text = match.group(1)
          text = text.strip()

          component = components.get(keyword, [])
          component.append(text)
          components[keyword] = component
          continue

        match = re.match('\\s*:(?:raises|raise)\\s+(\\w+)\\s*:(.*)?$', line)
        if match:
          keyword = 'Raises'
          exception = match.group(1)
          text = match.group(2)
          text = text.strip()

          component = components.get(keyword, [])
          component.append('- `{}`: {}'.format(exception, text))
          components[keyword] = component
          continue

      if keyword is not None:
        components[keyword].append(line)
      else:
        lines.append(line)

    for key in components:
      self._append_section(lines, key, components)

    section.content = '\n'.join(lines)

  @staticmethod
  def _append_section(lines, key, sections):
    section = sections.get(key)
    if not section:
      return

    if lines and lines[-1]:
      lines.append('')

    lines.extend(['**{}**:'.format(key), ''])  # add an extra line because of markdown syntax
    lines.extend(section)




import textwrap
from .utils import PreprocessorTest
from pydocmd.preprocessors.rst import Preprocessor


class RSTPreprocessorTest(PreprocessorTest):

  preprocessor = Preprocessor()

  code = textwrap.dedent('''
    def foo(a, b):
      """
      ```
      Param(foo, foo=bar)
      ```
      This is the main documentation!

      :param line1: The docs for line1
      :param line2: This docs for
      line2
      :return: This return
      can be multi-line!
      :raises ValueError: If something bad happens!
      """
  ''')

  expected_docs = textwrap.dedent('''
      ```
      Param(foo, foo=bar)
      ```
      This is the main documentation!

    **Arguments**:

    - `line1`: The docs for line1
    - `line2`: This docs for
      line2

    **Returns**:

    This return
      can be multi-line!

    **Raises**:

    - `ValueError`: If something bad happens!
  ''')



import pytest
import io

from pydocmd.document import Section


def test_preprocess_section():
  section = Section(None)
  section.depth = 1
  section.title = 'My Header'
  section.content = 'content'
  section.identifier = 'section-identifier'

  html_header_buffer = io.StringIO()
  section.render(html_header_buffer)
  assert html_header_buffer.getvalue() == "<h1 id=\"section-identifier\">My Header</h1>\n\ncontent\n"

  section.header_type = 'markdown'
  markdown_header_buffer = io.StringIO()
  section.render(markdown_header_buffer)
  assert markdown_header_buffer.getvalue() == "\n# My Header\ncontent\n"

  with pytest.raises(ValueError):
    section.header_type = 'invalid'
    section.render(markdown_header_buffer)



# Copyright (c) 2017  Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""
This module implements the structural representation of an API documentation
in separate documents and symbolic names. The final documentation is rendered
from this structured representation.
"""

from __future__ import print_function
import os


class Section(object):
  """
  A section represents a part of a #Document. It contains Markdown-formatted
  content that will be rendered into a file at some point.

  # Attributes
  doc (Document): The document that the section belongs to.
  identifier (str, None): The globally unique identifier of the section. This
    identifier usually matches the name of the element that the section
    describes (eg. a class or function) and will be used for cross-referencing.
  title (str, None): The title of the section. If specified, it will be
    rendered before `section.content` and the header-size will depend on
    the `section.depth`.
  depth (int): The depth of the section, defaults to 1. Currently only affects
    the header-size that is rendered for the `section.title`.
  content (str): The Markdown-formatted content of the section.
  """

  def __init__(self, doc, identifier=None, title=None, depth=1, content=None, header_type='html'):
    self.doc = doc
    self.identifier = identifier
    self.title = title
    self.depth = depth
    self.content = content if content is not None else '*Nothing to see here.*'
    self.header_type = header_type

  def render(self, stream):
    """
    Render the section into *stream*.
    """

    if self.header_type == 'html':
      print('<h{depth} id="{id}">{title}</h{depth}>\n'
          .format(depth=self.depth, id=self.identifier, title=self.title),
          file=stream)
    elif self.header_type == 'markdown':
      print('\n' + ('#' * self.depth), self.title, file=stream)
    else:
      raise ValueError('Invalid header type: %s' % self.header_type)
    print(self.content, file=stream)

  @property
  def index(self):
    """
    Returns the #Index that this section is associated with, accessed via
    `section.document`.
    """

    return self.document.index


class Document(object):
  """
  Represents a single document that may contain several #Section#s. Every
  document *must* have a relative URL associated with it.

  # Attributes
  index (Index): The index that the document belongs to.
  url (str): The relative URL of the document.
  """

  def __init__(self, index, url):
    self.index = index
    self.url = url
    self.sections = []


class Index(object):
  """
  The index manages all documents and sections globally. It keeps track of
  the symbolic names allocated for the sections to be able to link to them
  from other sections.

  # Attributes
  documents (dict):
  sections (dict):
  """

  def __init__(self):
    self.documents = {}
    self.sections = {}

  def new_document(self, filename, url=None):
    """
    Create a new document.

    # Arguments
    filename (str): The filename of the document. Must end with `.md`.
    url (str): The relative URL of the document. If omitted, will be
      automatically deduced from *filename* (same without the `.md` suffix).

    # Raises
    ValueError: If *filename* does not end with `.md`.
    ValueError: If *filename* is not a relative path.
    ValueError: If a document with the specified *filename* already exists.
    """

    if not filename.endswith('.md'):
      raise ValueError('filename must end with `.md`')
    if os.path.isabs(filename):
      raise ValueError('filename must be relative')
    if filename in self.documents:
      raise ValueError(
        'document filename {!r} already used'.format(filename))
    if not url:
      url = filename[:-3]

    doc = Document(self, url)
    self.documents[filename] = doc
    return doc

  def new_section(self, doc, *args, **kwargs):
    """
    Create a new section in the specified document. The arguments for this
    method match the parameters for the #Section constructor.

    # Raises
    ValueError: If the section identifier is already used.
    """

    section = Section(doc, *args, **kwargs)
    if section.identifier:
      if section.identifier in self.sections:
        raise ValueError('section identifier {!r} already used'
                 .format(section.identifier))
      self.sections[section.identifier] = section
    doc.sections.append(section)
    return section



import re


class Preprocessor:
  """
  This class implements the preprocessor for Google and PEP 257 docstrings.

  https://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html
  https://www.python.org/dev/peps/pep-0257/
  """
  _param_res = [
    re.compile(r'^(?P<param>\S+):\s+(?P<desc>.+)$'),
    re.compile(r'^(?P<param>\S+)\s+\((?P<type>[^)]+)\):\s+(?P<desc>.+)$'),
    re.compile(r'^(?P<param>\S+)\s+--\s+(?P<desc>.+)$'),
    re.compile(
      r'^(?P<param>\S+)\s+\{\[(?P<type>\S+)\]\}\s+--\s+(?P<desc>.+)$'),
    re.compile(
      r'^(?P<param>\S+)\s+\{(?P<type>\S+)\}\s+--\s+(?P<desc>.+)$'),
  ]

  _keywords_map = {
    'Args:': 'Arguments',
    'Arguments:': 'Arguments',
    'Attributes:': 'Attributes',
    'Example:': 'Examples',
    'Examples:': 'Examples',
    'Keyword Args:': 'Arguments',
    'Keyword Arguments:': 'Arguments',
    'Methods:': 'Methods',
    'Note:': 'Notes',
    'Notes:': 'Notes',
    'Other Parameters:': 'Arguments',
    'Parameters:': 'Arguments',
    'Return:': 'Returns',
    'Returns:': 'Returns',
    'Raises:': 'Raises',
    'References:': 'References',
    'See Also:': 'See Also',
    'Todo:': 'Todo',
    'Warning:': 'Warnings',
    'Warnings:': 'Warnings',
    'Warns:': 'Warns',
    'Yield:': 'Yields',
    'Yields:': 'Yields',
  }

  def __init__(self, config=None):
    self.config = config

  def get_section_names(self):
    return list(self._keywords_map.keys())

  def preprocess_section(self, section):
    """
    Preprocessors a given section into it's components.
    """
    lines = []
    in_codeblock = False
    keyword = None
    components = {}

    for line in section.content.split('\n'):
      line = line.strip()

      if line.startswith("```"):
        in_codeblock = not in_codeblock

      if in_codeblock:
        lines.append(line)
        continue

      if line in self._keywords_map:
        keyword = self._keywords_map[line]
        continue

      if keyword is None:
        lines.append(line)
        continue

      if keyword not in components:
        components[keyword] = []

      for param_re in self._param_res:
        param_match = param_re.match(line)
        if param_match:
          if 'type' in param_match.groupdict():
            components[keyword].append(
              '- `{param}` _{type}_ - {desc}'.format(**param_match.groupdict()))
          else:
            components[keyword].append(
              '- `{param}` - {desc}'.format(**param_match.groupdict()))
          break

      if not param_match:
        components[keyword].append('  {line}'.format(line=line))

    for key in components:
      self._append_section(lines, key, components)

    section.content = '\n'.join(lines)

  @staticmethod
  def _append_section(lines, key, sections):
    section = sections.get(key)
    if not section:
      return

    if lines and lines[-1]:
      lines.append('')

    # add an extra line because of markdown syntax
    lines.extend(['**{}**:'.format(key), ''])
    lines.extend(section)




import textwrap
from .utils import PreprocessorTest
from pydocmd.preprocessors.google import Preprocessor


class GooglePreprocessorTest(PreprocessorTest):

  # NOTE: The google preprocessor does not produce clean line endings.
  trim_trailing_whitespace = True

  preprocessor = Preprocessor()

  code = textwrap.dedent('''
    def foo(a, b):
      """Foo a and b.

      This is a function that computes a foo combination of two integers.

      Args:
        a (int): The integer A.
        b (int): The integer B.

      Returns:
        int: Foo combination of  ``a`` and ``b``.
      """
  ''')

  expected_docs = textwrap.dedent('''
    Foo a and b.

    This is a function that computes a foo combination of two integers.

    **Arguments**:

    - `a` _int_ - The integer A.
    - `b` _int_ - The integer B.


    **Returns**:

    - `int` - Foo combination of  ``a`` and ``b``.
  ''')
